@@ -1,18 +1,22 @@
 package Graph;
 
-# COMMENT THESE OUT FOR TESTING AND PRODUCTION.
-# $SIG{__DIE__ } = sub { use Carp; confess };
-# $SIG{__WARN__} = sub { use Carp; confess };
-
 use strict;
 
+BEGIN {
+    if (0) { # SET THIS TO ZERO FOR TESTING AND RELEASES!
+	$SIG{__DIE__ } = \&__carp_confess;
+	$SIG{__WARN__} = \&__carp_confess;
+    }
+    sub __carp_confess { require Carp; Carp::confess(@_) }
+}
+
 use Graph::AdjacencyMap qw(:flags :fields);
 
 use vars qw($VERSION);
 
-$VERSION = '0.65';
+$VERSION = '0.75';
 
-require 5.005;
+require 5.006; # Weak references are absolutely required.
 
 use Graph::AdjacencyMap::Heavy;
 use Graph::AdjacencyMap::Light;
@@ -55,19 +59,19 @@ sub _U () { 5 } # Union-Find.
 # - _UNORDUNIQ: just a union of _UNORD and UNIQ
 # Vertices are assumed to be _UNORDUNIQ; edges assume none of these flags.
 
-use Graph::Attribute array => _A, map => "graph";
+use Graph::Attribute array => _A, map => 'graph';
 
 sub _COMPAT02 () { 0x00000001 }
 
 sub stringify {
     my $g = shift;
     my $o = $g->is_undirected;
-    my $e = $o ? "=" : "-";
+    my $e = $o ? '=' : '-';
     my @e =
 	map {
 	    my @v =
 		map {
-		    ref($_) eq 'ARRAY' ? "[" . join(" ", @$_). "]" : "$_"
+		    ref($_) eq 'ARRAY' ? "[" . join(" ", @$_) . "]" : "$_"
 		}
 	    @$_;
 	    join($e, $o ? sort { "$a" cmp "$b" } @v : @v) } $g->edges05;
@@ -89,8 +93,6 @@ use overload
     'eq' => \&eq,
     'ne' => \&ne;
 
-require Data::Dumper; # for _dump
-
 sub _opt {
     my ($opt, $flags, %flags) = @_;
     while (my ($flag, $FLAG) = each %flags) {
@@ -268,7 +270,7 @@ sub new {
 
     $g->[ _F ] = $gflags;
     $g->[ _G ] = 0;
-    $g->[ _V ] = ($vflags & _HYPER) ?
+    $g->[ _V ] = ($vflags & (_HYPER | _MULTI)) ?
 	Graph::AdjacencyMap::Heavy->_new($uflags, 1) :
 	    (($vflags & ~_UNORD) ?
 	     Graph::AdjacencyMap::Vertex->_new($uflags, 1) :
@@ -347,7 +349,18 @@ sub add_vertex {
 	    Carp::croak("Graph::add_vertex: use add_vertices for more than one vertex or use hypervertexed");
 	}
 	for my $v ( @_ ) {
-	    $g->[ _V ]->set_path( $v ) unless $g->has_vertex( $v );
+	    if (defined $v) {
+		$g->[ _V ]->set_path( $v ) unless $g->has_vertex( $v );
+	    } else {
+		require Carp;
+		Carp::croak("Graph::add_vertex: undef vertex");
+	    }
+	}
+    }
+    for my $v ( @_ ) {
+	unless (defined $v) {
+	    require Carp;
+	    Carp::croak("Graph::add_vertex: undef vertex");
 	}
     }
     $g->[ _V ]->set_path( @_ );
@@ -589,7 +602,10 @@ sub add_vertex_by_id {
 sub add_vertex_get_id {
     my $g = shift;
     $g->expect_multivertexed;
-    return $g->[ _V ]->set_path_by_multi_id( @_, _GEN_ID );
+    my $id = $g->[ _V ]->set_path_by_multi_id( @_, _GEN_ID );
+    $g->[ _G ]++;
+    $g->_union_find_add_vertex( @_ ) if $g->has_union_find;
+    return $id;
 }
 
 sub has_vertex_by_id {
@@ -629,11 +645,12 @@ sub add_edge_by_id {
 
 sub add_edge_get_id {
     my $g = shift;
-    unless ($g->is_multiedged) {
-	require Carp;
-	Carp::croak("Graph::add_edge_get_id: graph not multiedged");
-    }
-    return $g->[ _E ]->set_path( $g->_add_edge( @_ ), _GEN_ID );
+    $g->expect_multiedged;
+    my @i = $g->_add_edge( @_ );
+    my $id = $g->[ _E ]->set_path_by_multi_id( @i, _GEN_ID );
+    $g->_union_find_add_edge( @i ) if $g->has_union_find;
+    $g->[ _G ]++;
+    return $id;
 }
 
 sub has_edge_by_id {
@@ -660,7 +677,9 @@ sub delete_edge_by_id {
 sub get_multiedge_ids {
     my $g = shift;
     $g->expect_multiedged;
-    $g->[ _E ]->get_multi_ids( $g->_vertex_ids( @_ ) );
+    my @id = $g->_vertex_ids( @_ );
+    return unless @id;
+    $g->[ _E ]->get_multi_ids( @id );
 }
 
 ###
@@ -741,11 +760,13 @@ sub _edges_from {
 	if (wantarray) {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if ($ev->[0] == $vi || $ev->[-1] == $vi) && !$ev{$ei}++;
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if $ev->[0] == $vi && !$ev{$ei}++;
 		}
@@ -753,11 +774,12 @@ sub _edges_from {
 	} else {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
-		    $en++
-			if ($ev->[0] == $vi || $ev->[-1] == $vi);
+		    next unless @$ev;
+		    $en++ if ($ev->[0] == $vi || $ev->[-1] == $vi);
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    $en++ if $ev->[0] == $vi;
 		}
 	    }
@@ -789,11 +811,13 @@ sub _edges_to {
 	if (wantarray) {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if ($ev->[-1] == $vi || $ev->[0] == $vi) && !$ev{$ei}++;
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    push @e, [ $ei, $ev ]
 			if $ev->[-1] == $vi && !$ev{$ei}++;
 		}
@@ -801,10 +825,12 @@ sub _edges_to {
 	} else {
 	    if ($o) {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    $en++ if $ev->[-1] == $vi || $ev->[0] == $vi;
 		}
 	    } else {
 		while (my ($ei, $ev) = each %{ $Ei }) {
+		    next unless @$ev;
 		    $en++ if $ev->[-1] == $vi;
 		}
 	    }
@@ -1150,7 +1176,7 @@ sub delete_cycle {
 
 sub has_cycle {
     my $g = shift;
-    $g->has_path(@_, $_[0]);
+    @_ ? ($g->has_path(@_, $_[0]) ? 1 : 0) : 0;
 }
 
 sub has_a_cycle {
@@ -1369,7 +1395,7 @@ sub set_edge_attribute_by_id {
     $g->expect_multiedged;
     my $value = pop;
     my $attr  = pop;
-    $g->add_edge_by_id( @_ ) unless $g->has_edge_by_id( @_ );
+    # $g->add_edge_by_id( @_ ) unless $g->has_edge_by_id( @_ );
     my $id = pop;
     $g->[ _E ]->_set_path_attr( $g->_vertex_ids( @_ ), $id, $attr, $value );
 }
@@ -1600,6 +1626,7 @@ sub copy {
 *copy_graph = \&copy;
 
 sub deep_copy {
+    require Data::Dumper;
     my $g = shift;
     my $d = Data::Dumper->new([$g]);
     use vars qw($VAR1);
@@ -2000,7 +2027,7 @@ sub _get_options {
 # Random constructors and accessors.
 #
 
-sub fisher_yates_shuffle (@) {
+sub __fisher_yates_shuffle (@) {
     # From perlfaq4, but modified to be non-modifying.
     my @a = @_;
     my $i = @a;
@@ -2014,9 +2041,14 @@ sub fisher_yates_shuffle (@) {
 BEGIN {
     sub _shuffle(@);
     # Workaround for the Perl bug [perl #32383] where -d:Dprof and
-    # List::Util::Shuffle do not like each other: if any debugging
-    # (-d) flags are on, fall back to the Fisher-Yates shuffle.
-    *_shuffle = $^P ? \&fisher_yates_shuffle : \&List::Util::shuffle;
+    # List::Util::shuffle do not like each other: if any debugging
+    # (-d) flags are on, fall back to our own Fisher-Yates shuffle.
+    # The bug was fixed by perl changes #26054 and #26062, which
+    # went to Perl 5.9.3.  If someone tests this with a pre-5.9.3
+    # bleadperl that calls itself 5.9.3 but doesn't yet have the
+    # patches, oh, well.
+    *_shuffle = $^P && $] < 5.009003 ?
+	\&__fisher_yates_shuffle : \&List::Util::shuffle;
 }
 
 sub random_graph {
@@ -2166,7 +2198,7 @@ sub MST_Kruskal {
 }
 
 sub _MST_add {
-    my ($g, $HF, $r, $attr, $unseen) = @_;
+    my ($g, $h, $HF, $r, $attr, $unseen) = @_;
     for my $s ( grep { exists $unseen->{ $_ } } $g->successors( $r ) ) {
 	$HF->add( Graph::MSTHeapElem->new( $r, $s, $g->get_edge_attribute( $r, $s, $attr ) ) );
     }
@@ -2213,22 +2245,25 @@ sub _root_opt {
 }
 
 sub _heap_walk {
-    my ($g, $h, $add, $etc) = splice @_, 0, 4;
+    my ($g, $h, $add, $etc) = splice @_, 0, 4; # Leave %opt in @_.
 
     my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) = $g->_root_opt(@_);
     my $HF = Heap::Fibonacci->new;
 
     while (defined $r) {
-	$add->($g, $HF, $r, $attr, $unseenh, $etc);
+	# print "r = $r\n";
+	$add->($g, $h, $HF, $r, $attr, $unseenh, $etc);
 	delete $unseenh->{ $r };
 	while (defined $HF->top) {
 	    my $t = $HF->extract_top;
+	    # use Data::Dumper; print "t = ", Dumper($t);
 	    if (defined $t) {
 		my ($u, $v, $w) = $t->val;
+		# print "extracted top: $u $v $w\n";
 		if (exists $unseenh->{ $v }) {
 		    $h->set_edge_attribute($u, $v, $attr, $w);
 		    delete $unseenh->{ $v };
-		    $add->($g, $HF, $v, $attr, $unseenh, $etc);
+		    $add->($g, $h, $HF, $v, $attr, $unseenh, $etc);
 		}
 	    }
 	}
@@ -2517,13 +2552,14 @@ sub _connected_components {
 		@cci{ keys %icci } = values %icci;
 	    }
 	} else {
-	    my %r; @r{ $g->unique_vertices } = ();
+	    my @u = $g->unique_vertices;
+	    my %r; @r{ @u } = @u;
 	    my $froot = sub {
-		wantarray ? ( keys %r ? scalar each %r: ()) : each %r;
+		(each %r)[1];
 	    };
 	    my $nroot = sub {
-		$cc++;
-		wantarray ? ( keys %r ? scalar each %r: ()) : each %r;
+		$cc++ if keys %r;
+		(each %r)[1];
 	    };
 	    my $t = Graph::Traversal::DFS->new($g,
 					       first_root => $froot,
@@ -2592,17 +2628,23 @@ sub same_connected_components {
     }
 }
 
-my $super_component = sub { join("+", sort @{ $_[0] }) };
+my $super_component = sub { join("+", sort @_) };
 
 sub connected_graph {
-    my $g = shift;
+    my ($g, %opt) = @_;
     $g->expect_undirected;
     my $cg = Graph->new(undirected => 1);
     if ($g->has_union_find && $g->vertices == 1) {
+	# TODO: super_component?
 	$cg->add_vertices($g->vertices);
     } else {
+	my $sc_cb =
+	    exists $opt{super_component} ?
+		$opt{super_component} : $super_component;
 	for my $cc ( $g->connected_components() ) {
-	    $cg->add_vertex($super_component->($cc));
+	    my $sc = $sc_cb->(@$cc);
+	    $cg->add_vertex($sc);
+	    $cg->set_vertex_attribute($sc, 'subvertices', [ @$cc ]);
 	}
     }
     return $cg;
@@ -2805,18 +2847,28 @@ sub strongly_connected_graph {
 
     $u->dfs;
 
-    my $hypervertex = $super_component;
+    my $sc_cb;
+    my $hv_cb;
 
-    _opt_get(\%attr, hypervertex => \$hypervertex);
+    _opt_get(\%attr, super_component => \$sc_cb);
+    _opt_get(\%attr, hypervertex => \$hv_cb);
     _opt_unknown(\%attr);
 
+    if (defined $hv_cb && !defined $sc_cb) {
+	$sc_cb = sub { $hv_cb->( [ @_ ] ) };
+    }
+    unless (defined $sc_cb) {
+	$sc_cb = $super_component;
+    }
+
     my $s = Graph->new;
 
     my %c;
     my @s;
     for (my $i = 0; $i <  @c; $i++) {
 	my $c = $c[$i];
-	$s->add_vertex( $s[$i] = $hypervertex->($c) );
+	$s->add_vertex( $s[$i] = $sc_cb->(@$c) );
+	$s->set_vertex_attribute($s[$i], 'subvertices', [ @$c ]);
 	for my $v (@$c) {
 	    $c{$v} = $i;
 	}
@@ -2855,10 +2907,10 @@ sub _make_bcc {
     my %b;
     while (@$S) {
 	my $t = pop @$S;
-	$b{ $t }++;
+	$b{ $t } = $t;
 	last if $t eq $v;
     }
-    return [ keys %b, $c ];
+    return [ values %b, $c ];
 }
 
 sub biconnectivity {
@@ -2889,84 +2941,133 @@ sub biconnectivity {
 	my %L = ( $r => 1 );
 	my @S = ( $r );
 	my %A;
-	for my $v ($g->vertices) {
-	    @{ $A{ $v } }{ $g->successors( $v ) } = ();
+	my @V = $g->vertices;
+
+	# print "V : @V\n";
+	# print "r : $r\n";
+
+	my %T; @T{ @V } = @V;
+	my %S; # Self-loops.
+
+	for my $w (@V) {
+	    my @s = grep {
+		if ($_ eq $w) {
+		    $S{$w} = $w;
+		    delete $T{$_};
+		    0;
+		} else {
+		    1;
+		}
+	    } $g->successors( $w );
+	    @{ $A{ $w } }{ @s } = @s;
 	}
-	my @C;
-	my $Avok;
-	do {
-	    my $w;
-	    do {
-		$w = first { !$U{ $v }{ $_ } } _shuffle keys %{ $A{ $v } };
-		if (defined $w) {
-		    $U{ $v }{ $w }++;
-		    $U{ $w }{ $v }++;
-		    if ($I{ $w } == 0) {
-			$P{ $w } = $v;
-			$i++;
-			$I{ $w } = $i;
-			$L{ $w } = $i;
-			push @S, $w;
-			$v = $w;
-		    } else {
-			$L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
-		    }
-		}
-	    } while (defined $w);
-	    if (!defined $P{ $v }) {
-		# Do nothing.
-	    } elsif ($P{ $v } ne $r) {
-		if ($L{ $v } < $I{ $P{ $v } }) {
-		    $L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
-		} else {
-		    $AP{ $P{ $v } }++;
-		    push @C, _make_bcc(\@S, $v, $P{ $v } );
-		}
-	    } else {
-		my $e;
-		for my $w (_shuffle keys %{ $A{ $r } }) {
-		    unless ($U{ $r }{ $w }) {
-			$e = $r;
-			last;
-		    }
-		}
-		$AP{ $e }++ if defined $e;
-		push @C, _make_bcc(\@S, $v, $r);
-	    }
-	    $v = defined $P{ $v } ? $P{ $v } : $r;
-	    $Avok = 0;
-	    if (defined $v) {
-		$Avok = 0;
-		if (keys %{ $A{ $v } }) {
-		    if (!exists $P{ $v }) {
-			for my $w (keys %{ $A{ $v } }) {
-			    $Avok++ if $U{ $v }{ $w };
-			}
-			$Avok = 0 unless $Avok == keys %{ $A{ $v } };
-		    }
-		} else {
-		    $Avok = 1;
-		}
-	    }
-	} until ($Avok);
+
+	# use Data::Dumper;
+	# print "T : ", Dumper(\%T);
+	# print "A : ", Dumper(\%A);
 
 	my %V2BC;
-	for (my $i = 0; $i < @C; $i++) {
-	    for my $v (@{ $C[ $i ]}) {
-		$V2BC{ $v }{ $i }++;
+	my @BR;
+	my @BC;
+
+	my @C;
+	my $Avok;
+
+	while (keys %T) {
+	    # print "T = ", Dumper(\%T);
+	    do {
+		my $w;
+	        do {
+		    my @w = _shuffle values %{ $A{ $v } };
+		    # print "w = @w\n";
+		    $w = first { !$U{ $v }{ $_ } } @w;
+		    if (defined $w) {
+			# print "w = $w\n";
+			$U{ $v }{ $w }++;
+			$U{ $w }{ $v }++;
+			if ($I{ $w } == 0) {
+			    $P{ $w } = $v;
+			    $i++;
+			    $I{ $w } = $i;
+			    $L{ $w } = $i;
+			    push @S, $w;
+			    $v = $w;
+			} else {
+			    $L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
+			}
+		    }
+		} while (defined $w);
+		# print "U = ", Dumper(\%U);
+		# print "P = ", Dumper(\%P);
+		# print "L = ", Dumper(\%L);
+		if (!defined $P{ $v }) {
+		    # Do nothing.
+		} elsif ($P{ $v } ne $r) {
+		    if ($L{ $v } < $I{ $P{ $v } }) {
+			$L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
+		    } else {
+			$AP{ $P{ $v } } = $P{ $v };
+			push @C, _make_bcc(\@S, $v, $P{ $v } );
+		    }
+		} else {
+		    my $e;
+		    for my $w (_shuffle keys %{ $A{ $r } }) {
+			# print "w = $w\n";
+			unless ($U{ $r }{ $w }) {
+			    $e = $r;
+			    # print "e = $e\n";
+			    last;
+			}
+		    }
+		    $AP{ $e } = $e if defined $e;
+		    push @C, _make_bcc(\@S, $v, $r);
+		}
+		# print "AP = ", Dumper(\%AP);
+		# print "C  = ", Dumper(\@C);
+		# print "L  = ", Dumper(\%L);
+		$v = defined $P{ $v } ? $P{ $v } : $r;
+		# print "v = $v\n";
+		$Avok = 0;
+		if (defined $v) {
+		    if (keys %{ $A{ $v } }) {
+			if (!exists $P{ $v }) {
+			    for my $w (keys %{ $A{ $v } }) {
+				$Avok++ if $U{ $v }{ $w };
+			    }
+			    # print "Avok/1 = $Avok\n";
+			    $Avok = 0 unless $Avok == keys %{ $A{ $v } };
+			    # print "Avok/2 = $Avok\n";
+			}
+		    } else {
+			$Avok = 1;
+			# print "Avok/3 = $Avok\n";
+		    }
+		}
+	    } until ($Avok);
+
+	    last if @C == 0 && !exists $S{$v};
+
+	    for (my $i = 0; $i < @C; $i++) {
+		for my $v (@{ $C[ $i ]}) {
+		    $V2BC{ $v }{ $i }++;
+		    delete $T{ $v };
+		}
+	    }
+
+	    for (my $i = 0; $i < @C; $i++) {
+		if (@{ $C[ $i ] } == 2) {
+		    push @BR, $C[ $i ];
+		} else {
+		    push @BC, $C[ $i ];
+		}
+	    }
+
+	    if (keys %T) {
+		$r = $v = each %T;
 	    }
 	}
 
-	my @BR;
-	my @BC;
-	for (my $i = 0; $i < @C; $i++) {
-	    if (@{ $C[ $i ] } == 2) {
-		push @BR, $C[ $i ];
-	    } else {
-		push @BC, $C[ $i ];
-	    }
-	}
-	$bcc = [ [keys %AP], \@BC, \@BR, \%V2BC ];
+	$bcc = [ [values %AP], \@BC, \@BR, \%V2BC ];
 	$g->set_graph_attribute('_bcc', [ $g->[ _G ], $bcc ]);
     }
 
@@ -3042,11 +3143,15 @@ sub same_biconnected_components {
 }
 
 sub biconnected_graph {
-    my $g = shift;
+    my ($g, %opt) = @_;
     my ($bc, $v2bc) = ($g->biconnectivity)[1, 3];
     my $bcg = Graph::Undirected->new;
+    my $sc_cb =
+	exists $opt{super_component} ?
+	    $opt{super_component} : $super_component;
     for my $c (@$bc) {
-	$bcg->add_vertex($super_component->($c));
+	$bcg->add_vertex(my $s = $sc_cb->(@$c));
+	$bcg->set_vertex_attribute($s, 'subvertices', [ @$c ]);
     }
     my %k;
     for my $i (0..$#$bc) {
@@ -3059,8 +3164,8 @@ sub biconnected_graph {
 		for my $u (@u) {
 		    if (exists $j{ $u }) {
 			unless ($k{ $i }{ $j }++) {
-			    $bcg->add_edge($super_component->($bc->[$i]),
-					   $super_component->($bc->[$j]));
+			    $bcg->add_edge($sc_cb->(@{$bc->[$i]}),
+					   $sc_cb->(@{$bc->[$j]}));
 			}
 			last;
 		    }
@@ -3082,33 +3187,67 @@ sub bridges {
 #
 
 sub _SPT_add {
-    my ($g, $HF, $r, $attr, $unseen, $etc) = @_;
-    for my $s ( grep { $unseen->{ $_ } } $g->successors( $r ) ) {
-	my $t = $g->get_edge_attribute( $r, $s, $attr ) || 1;
+    my ($g, $h, $HF, $r, $attr, $unseen, $etc) = @_;
+    my $etc_r = $etc->{ $r } || 0;
+    for my $s ( grep { exists $unseen->{ $_ } } $g->successors( $r ) ) {
+	my $t = $g->get_edge_attribute( $r, $s, $attr );
+	$t = 1 unless defined $t;
 	if ($t < 0) {
 	    require Carp;
 	    Carp::croak("Graph::SPT_Dijkstra: edge $r-$s is negative ($t)");
 	}
-	if (!defined($etc->{ $s }) || $etc->{ $r } + $t < $etc->{ $s }) {
-	    $etc->{ $s } = ($etc->{ $r } || 0) + $t;
-	    $HF->add( Graph::SPTHeapElem->new($r, $s, $t) );
+	if (!defined($etc->{ $s }) || ($etc_r + $t) < $etc->{ $s }) {
+	    my $etc_s = $etc->{ $s } || 0;
+	    $etc->{ $s } = $etc_r + $t;
+	    # print "$r - $s : setting $s to $etc->{ $s } ($etc_r, $etc_s)\n";
+	    $h->set_vertex_attribute( $s, $attr, $etc->{ $s });
+	    $h->set_vertex_attribute( $s, 'p', $r );
+	    $HF->add( Graph::SPTHeapElem->new($r, $s, $etc->{ $s }) );
 	}
     }
 }
 
 sub SPT_Dijkstra {
     my $g = shift;
-    my %etc;
-    $g->_heap_walk($g->new(), \&_SPT_add, \%etc, @_);
+    my %opt = @_ == 1 ? (first_root => $_[0]) : @_;
+    my $first_root = $opt{ first_root };
+    unless (defined $first_root) {
+	$opt{ first_root } = $first_root = $g->random_vertex();
+    }
+    my $spt_di = $g->get_graph_attribute('_spt_di');
+    unless (defined $spt_di && exists $spt_di->{ $first_root } && $spt_di->{ $first_root }->[ 0 ] == $g->[ _G ]) {
+	my %etc;
+	my $sptg = $g->_heap_walk($g->new(), \&_SPT_add, \%etc, %opt);
+	$spt_di->{ $first_root } = [ $g->[ _G ], $sptg ];
+	$g->set_graph_attribute('_spt_di', $spt_di);
+    }
+    return $spt_di->{ $first_root }->[ 1 ];
 }
 
 *SSSP_Dijkstra = \&SPT_Dijkstra;
 
 *single_source_shortest_paths = \&SPT_Dijkstra;
 
-sub SPT_Bellman_Ford {
-    my $g = shift;
-    my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) = $g->_root_opt(@_);
+sub SP_Dijkstra {
+    my ($g, $u, $v) = @_;
+    my $sptg = $g->SPT_Dijkstra(first_root => $u);
+    my @path = ($v);
+    my %seen;
+    my $V = $g->vertices;
+    my $p;
+    while (defined($p = $sptg->get_vertex_attribute($v, 'p'))) {
+	last if exists $seen{$p};
+	push @path, $p;
+	$v = $p;
+	$seen{$p}++;
+	last if keys %seen == $V || $u eq $v;
+    }
+    # @path = () if @path && $path[-1] ne $u;
+    return reverse @path;
+}
+
+sub _SPT_Bellman_Ford {
+    my ($g, $opt, $unseenh, $unseena, $r, $next, $code, $attr) = @_;
     my %d;
     $d{ $r } = 0;
     my %p;
@@ -3153,17 +3292,57 @@ sub SPT_Bellman_Ford {
 	}
     }
 
-    my $h = $g->new();
-    for my $v (keys %p) {
-	$h->add_edge( $p{ $v }, $v );
-#	$h->set_vertex_attribute( $v, $attr, $d{ $v } );
+    return (\%p, \%d );
+}
+
+sub SPT_Bellman_Ford {
+    my $g = shift;
+
+    my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) = $g->_root_opt(@_);
+
+    unless (defined $r) {
+	$r = $g->random_vertex();
     }
 
-    return $h;
+    my $spt_bf = $g->get_graph_attribute('_spt_bf');
+    unless (defined $spt_bf && exists $spt_bf->{ $r } && $spt_bf->{ $r }->[ 0 ] == $g->[ _G ]) {
+	my ($p, $d) = $g->_SPT_Bellman_Ford($opt, $unseenh, $unseena, $r, $next, $code, $attr);
+	my $h = $g->new();
+	for my $v (keys %$p) {
+	    my $u = $p->{ $v };
+	    $h->add_edge( $u, $v );
+	    $h->set_edge_attribute( $u, $v, $attr,
+				    $g->get_edge_attribute($u, $v, $attr));
+	    $h->set_vertex_attribute( $v, $attr, $d->{ $v } );
+	    $h->set_vertex_attribute( $v, 'p', $u );
+	}
+	$spt_bf->{ $r } = [ $g->[ _G ], $h ];
+	$g->set_graph_attribute('_spt_bf', $spt_bf);
+    }
+
+    return $spt_bf->{ $r }->[ 1 ];
 }
 
 *SSSP_Bellman_Ford = \&SPT_Bellman_Ford;
 
+sub SP_Bellman_Ford {
+    my ($g, $u, $v) = @_;
+    my $sptg = $g->SPT_Dijkstra(first_root => $u);
+    my @path = ($v);
+    my %seen;
+    my $V = $g->vertices;
+    my $p;
+    while (defined($p = $sptg->get_vertex_attribute($v, 'p'))) {
+	last if exists $seen{$p};
+	push @path, $p;
+	$v = $p;
+	$seen{$p}++;
+	last if keys %seen == $V;
+    }
+    # @path = () if @path && "$path[-1]" ne "$u";
+    return reverse @path;
+}
+
 ###
 # Transitive Closure.
 #
@@ -3190,7 +3369,7 @@ sub transitive_closure_matrix {
     my $g = shift;
     my $tcm = $g->get_graph_attribute('_tcm');
     unless (defined $tcm) {
-	my $apsp = $g->APSP_Floyd_Warshall();
+	my $apsp = $g->APSP_Floyd_Warshall(@_);
 	$tcm = $apsp->get_graph_attribute('_tcm');
 	$g->set_graph_attribute('_tcm', $tcm);
     }
@@ -3224,7 +3403,7 @@ sub is_reachable {
 sub for_shortest_paths {
     my $g = shift;
     my $c = shift;
-    my $t = Graph::TransitiveClosure->new($g, path => 1);
+    my $t = $g->transitive_closure_matrix;
     my @v = $g->vertices;
     my $n = 0;
     for my $u (@v) {
@@ -3240,14 +3419,15 @@ sub for_shortest_paths {
 sub _minmax_path {
     my $g = shift;
     my $min;
-    my $max = 0;
+    my $max;
     my $minp;
     my $maxp;
     $g->for_shortest_paths(sub {
 			       my ($t, $u, $v, $n) = @_;
 			       my $l = $t->path_length($u, $v);
+			       return unless defined $l;
 			       my $p;
-			       if ($l > $max) {
+			       if ($u ne $v && (!defined $max || $l > $max)) {
 				   $max = $l;
 				   $maxp = $p = [ $t->path_vertices($u, $v) ];
 			       }
@@ -3259,24 +3439,129 @@ sub _minmax_path {
     return ($min, $max, $minp, $maxp);
 }
 
-sub graph_diameter {
-    my $g = shift;
-    my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
-    return $max;
-}
-
-sub longest_path {
+sub diameter {
     my $g = shift;
     my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
     return defined $maxp ? (wantarray ? @$maxp : $max) : undef;
 }
 
+*graph_diameter = \&diameter;
+
+sub longest_path {
+    my ($g, $u, $v) = @_;
+    my $t = $g->transitive_closure_matrix;
+    if (defined $u) {
+	if (defined $v) {
+	    return wantarray ?
+		$t->path_vertices($u, $v) : $t->path_length($u, $v);
+	} else {
+	    my $max;
+	    my @max;
+	    for my $v ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $max || $l > $max)) {
+		    $max = $l;
+		    @max = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @max : $max;
+	}
+    } else {
+	if (defined $v) {
+	    my $max;
+	    my @max;
+	    for my $u ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $max || $l > $max)) {
+		    $max = $l;
+		    @max = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @max : @max - 1;
+	} else {
+	    my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
+	    return defined $maxp ? (wantarray ? @$maxp : $max) : undef;
+	}
+    }
+}
+
+sub vertex_eccentricity {
+    my ($g, $u) = @_;
+    my $max;
+    for my $v ($g->vertices) {
+	next if $u eq $v;
+	my $l = $g->path_length($u, $v);
+	if (defined $l && (!defined $max || $l > $max)) {
+	    $max = $l;
+	}
+    }
+    return $max;
+}
+
 sub shortest_path {
+    my ($g, $u, $v) = @_;
+    my $t = $g->transitive_closure_matrix;
+    if (defined $u) {
+	if (defined $v) {
+	    return wantarray ?
+		$t->path_vertices($u, $v) : $t->path_length($u, $v);
+	} else {
+	    my $min;
+	    my @min;
+	    for my $v ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $min || $l < $min)) {
+		    $min = $l;
+		    @min = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @min : $min;
+	}
+    } else {
+	if (defined $v) {
+	    my $min;
+	    my @min;
+	    for my $u ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $min || $l < $min)) {
+		    $min = $l;
+		    @min = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @min : $min;
+	} else {
+	    my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
+	    return defined $minp ? (wantarray ? @$minp : $min) : undef;
+	}
+    }
+}
+
+sub radius {
     my $g = shift;
     my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
     return defined $minp ? (wantarray ? @$minp : $min) : undef;
 }
 
+sub center_vertices {
+    my ($g, $delta) = @_;
+    $delta = 0 unless defined $delta;
+    $delta = abs($delta);
+    my @c;
+    my $r = $g->radius;
+    if (defined $r) {
+	for my $v ($g->vertices) {
+	    my $e = $g->vertex_eccentricity($v);
+	    next unless defined $e;
+	    push @c, $v if abs($e - $r) <= $delta;
+	}
+    }
+    return @c;
+}
+
 sub average_path_length {
     my $g = shift;
     my @A = @_;
@@ -3411,6 +3696,7 @@ sub could_be_isomorphic {
 #
 
 sub _dump {
+    require Data::Dumper;
     my $d = Data::Dumper->new([$_[0]],[ref $_[0]]);
     defined wantarray ? $d->Dump : print $d->Dump;
 }
